#
#   5B - Editor
#

from pygame.locals import *
from pygame import draw
from albow.widget import Widget
from albow.controls import Label, Button, ImageButton, Row, Column, Frame
from albow.palette_view import PaletteView
from albow.resource import get_font, get_image, get_cursor
from controls import TextDisplay, TimeDisplay, text_to_time, time_to_text, \
	Form
from screen import Screen, menu_button
from level_view import LevelView
from part import Part
from wire import Wire, WirePoint, get_wire_color, num_wire_colors, wire_width
from vector_algebra import Point2, IdentTran2, ScaleTran2, Zero2
import parts, settings

toolbar_font = get_font(12, "Vera.ttf")
tool_palette_cell_size = (30, 30)
wire_palette_width = 50
part_palette_width = 120
wire_palette_item_height = 16
wire_palette_cols = 1
grid_size = 5

SELECT_TOOL = 0
ADD_TOOL = 1
WIRE_TOOL = 2
CUT_TOOL = 3

#----------------------------------------------------------------------------

class EditorMenuScreen(Screen):

	def __init__(self, shell):
		Screen.__init__(self, shell)
		game = shell.game
		title = Label("Level Editor", font = get_font(24, "VeraBd.ttf"))
		def mb(title, action, enable = None):
			b = menu_button(title, action = action, enable = enable)
			b.fg_color = (255, 128, 0)
			return b
		items = [
			mb("New Level Set...", shell.new_level_set),
			mb("New Level",        shell.edit_new_level),
			mb("Edit Level",       shell.show_editor_level, game.level_present),
			mb("Load Level...",    shell.load_level),
			mb("Save Level",       shell.save_level,        game.level_present),
			mb("Save Level As...", shell.save_level_as,     game.level_present),
			mb("Exit Editor",      shell.show_menu),
		]
		menu = Column(items, align = 'l')
		box = Column([title, menu], align = 'l', spacing = 24)
		self.center(box)
	
	def key_down(self, e):
		if e.cmd:
			self.cmd_key(e)
		else:
			Screen.key_down(self, e)
	
	def cmd_key(self, e):
		key = e.key
		if key == K_n:
			self.shell.edit_new_level()
		elif key == K_e:
			self.shell.show_editor_level()
		elif key == K_o:
			self.shell.load_level()
		elif key == K_s:
			if e.shift:
				self.shell.save_level_as()
			else:
				self.shell.save_level()
		elif key == K_q:
			self.shell.show_menu()
		else:
			Screen.key_down(self, e)
	
	def exit_editor(self):
		level = self.game_current_level
		if level:
			level.testing = False
		self.shell.show_menu()

#----------------------------------------------------------------------------

def lf(label, w):
	l = Label(label, font = toolbar_font)
	return Row([l, w])

class EditorLevelScreen(Screen):

	def __init__(self, shell):
		Screen.__init__(self, shell)
		sw, sh = settings.screen_size
		level_view = EditorLevelView(Rect((0, 0), (sw, sh)), shell)
		tool_palette = EditorToolPalette(level_view)
		self.wire_palette = WireColorPalette(level_view)
		wire_frame = self.wire_palette
		wire_popup = EditorWirePopup(wire_palette_width, wire_frame, level_view)
		self.part_palette = EditorPartPalette(level_view)
		part_frame = self.part_palette
		part_popup = EditorPartPopup(part_palette_width, part_frame, level_view)
		buttons = self.make_buttons(level_view)
		start_time_field = TimeDisplay(level_view, 'start_time',
			click_action = level_view.edit_start_time, font = toolbar_font)
		deadline_field = TimeDisplay(level_view, 'deadline',
			click_action = level_view.edit_deadline, font = toolbar_font)
		script_field = TextDisplay(100, level_view, 'script',
			click_action = level_view.edit_script, font = toolbar_font)
		change_indicator = ChangeIndicator(shell.game)
		box = Row(
			[
				tool_palette, 
				wire_popup, 
				part_popup,
				buttons, 
				lf("Init", start_time_field),
				lf("Dead", deadline_field),
				lf("Script", script_field),
				change_indicator,
			])
		#print "editor: toolbar height =", box.rect.height ###
		box.rect.bottomleft = (0, sh)
		self.editing_controls = box
		level_view.rect.height = box.rect.top
		wire_frame.rect.bottomleft = wire_popup.rect.left, box.rect.top
		part_frame.rect.bottomleft = part_popup.rect.left, box.rect.top
		self.add(level_view)
		self.add(wire_frame)
		self.add(part_frame)
		self.add(box)
		level_view.focus()
	
	def make_buttons(self, level_view):
		return Row([
			ImageButton("buttons/flip-h.png", action = level_view.flip_h),
			ImageButton("buttons/flip-v.png", action = level_view.flip_v),
			ImageButton("buttons/rotate.png", action = level_view.rotate),
			ImageButton("buttons/params.png", action = level_view.edit_params),
			ImageButton("buttons/test.png", action = level_view.test_level),
		], spacing = 0)

#----------------------------------------------------------------------------

class EditorToolPalette(PaletteView):

	images = [
		get_image("tools/select.png"),
		get_image("tools/add.png"),
		get_image("tools/wire.png"),
		get_image("tools/cut.png"),
	]
	
	highlight_style = 'fill'
	
	def __init__(self, editor):
		PaletteView.__init__(self, tool_palette_cell_size, 1, self.num_items())
		self.editor = editor
	
	def num_items(self):
		return len(self.images)
	
	def draw_item(self, surf, item_no, frame):
		image = self.images[item_no]
		r = image.get_rect()
		r.center = frame.center
		surf.blit(image, r)
	
	def click_item(self, item_no, e):
		self.editor.set_tool(item_no)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.tool

#----------------------------------------------------------------------------

class EditorPopupButton(Widget):

	font = toolbar_font
	fg_color = (255, 255, 255)
	border_width = 1
	margin = 3

	def __init__(self, width, client, editor):
		self.client = client
		self.editor = editor
		client.visible = False
		w, h = self.font.size("X")
		d = 2 * self.margin
		Widget.__init__(self, Rect(0, 0, width + d, h + d))

	def draw(self, surf):
		frame = surf.get_rect()
		self.draw_selection(surf, frame)
	
	def mouse_down(self, e):
		client = self.client
		client.visible = not client.visible

#----------------------------------------------------------------------------

class EditorPartPopup(EditorPopupButton):

	def draw_selection(self, surf, frame):
		name = parts.catalog[self.editor.part_no].name
		buf = self.font.render(name, True, self.fg_color)
		r = buf.get_rect()
		x, y = frame.midleft
		r.midleft = x + 3, y
		surf.blit(buf, r)
	
#----------------------------------------------------------------------------

class EditorWirePopup(EditorPopupButton):

	def draw_selection(self, surf, frame):
		color = get_wire_color(self.editor.wire_color_no)
		x, y = frame.topleft
		d = frame.height // 2 - 1
		draw.line(surf, color,
			(x + d, y + d),
			(x + frame.width - d, y + d),
			wire_width)

#----------------------------------------------------------------------------

class EditorPopupPalette(PaletteView):

	border_width = 1
	margin = 3

	def click_item(self, item_no, e):
		self.select_item(item_no)
		self.visible = False

#----------------------------------------------------------------------------

class EditorPartPalette(EditorPopupPalette):

	font = get_font(12, "Vera.ttf")
	fg_color = (255, 255, 255)
	bg_color = (0, 0, 0)
	highlight_style = 'fill'

	def __init__(self, editor):
		cw = part_palette_width
		ch = self.font.get_linesize()
		PaletteView.__init__(self, (cw, ch), self.num_items(), 1)
		self.editor = editor
	
	def num_items(self):
		return len(parts.catalog)
	
	def draw_item(self, surf, item_no, frame):
		name = parts.catalog[item_no].name
		buf = self.font.render(name, True, self.fg_color)
		surf.blit(buf, frame)
	
	def select_item(self, item_no):
		self.editor.part_no = item_no
		self.editor.set_tool(ADD_TOOL)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.part_no

#----------------------------------------------------------------------------

class WireColorPalette(EditorPopupPalette):

	def __init__(self, editor):
		nc = wire_palette_cols
		nr = (self.num_items() + nc - 1) // nc
		cw = wire_palette_width // nc
		ch = wire_palette_item_height
		PaletteView.__init__(self, (cw, ch), nr, nc)
		self.editor = editor
	
	def num_items(self):
		return num_wire_colors()
		
	def draw_item(self, surf, item_no, frame):
		x, y = frame.topleft
		d = frame.height // 2 - 1
		draw.line(surf, get_wire_color(item_no),
			(x + d, y + d),
			(x + frame.width - d, y + d),
			wire_width)

	def select_item(self, item_no):
		self.editor.wire_color_no = item_no
		self.editor.set_tool(WIRE_TOOL)
	
	def item_is_selected(self, item_no):
		return item_no == self.editor.wire_color_no

#----------------------------------------------------------------------------

class ChangeIndicator(Widget):

	fg_color = (255, 255, 255)
	dot_color = (255, 0, 0)

	def __init__(self, game):
		self.game = game
		Widget.__init__(self, Rect(0, 0, 10, 10))
	
	def draw(self, surf):
		frame = surf.get_rect()
		draw.rect(surf, self.fg_color, frame, 1)
		if self.game.current_level.changed:
			r = Rect(0, 0, 6, 6)
			r.center = frame.center
			surf.fill(self.dot_color, r)

#----------------------------------------------------------------------------

class EditorLevelView(LevelView):
	#  tool                 int
	#  part_no              int
	#  wire_color_no        int
	#  click_point          Point2
	#  rubber_point         Point2
	#  editing_wire_point   WirePoint or None
	#  dragged              boolean

	highlight_color = (0, 128, 255)
	add_cursor = get_cursor("cursors/add.png")
	solder_cursor = get_cursor("cursors/solder.png")
	wire_cursor = get_cursor("cursors/wire.png")
	cut_cursor = get_cursor("cursors/cutters.png")
	cut_wire_cursor = get_cursor("cursors/cutters-open.png")
	delete_wire_cursor = get_cursor("cursors/cutters-delete.png")

	def __init__(self, rect, shell):
		LevelView.__init__(self, rect, shell.game)
		self.shell = shell
		self.tool = SELECT_TOOL
		self.part_no = 0
		self.wire_color_no = 0
		self.click_point = None
		self.rubber_point = None
		self.editing_wire_point = None
		self.dragged = False

	level = property(lambda self: self.game.current_level)
	
	def set_tool(self, tool):
		if tool <> self.tool:
			if self.tool == WIRE_TOOL:
				self.finish_wire()
		self.tool = tool
		if tool <> SELECT_TOOL:
			self.game.current_level.clear_selection()
	
	def draw_with_transform(self, surf, t, with_selection):
		LevelView.draw_with_transform(self, surf, t, with_selection = True)
		level = self.level
		for item in level.selection:
			if isinstance(item, WirePoint):
				item.draw_highlight(surf, t, self.highlight_color)
		if self.tool == WIRE_TOOL:
			wp = self.editing_wire_point
			if wp:
				p = self.rubber_point
				if p:
					wp.draw_rubber(surf, t, p)
	
	def mouse_down(self, e):
		self.dragged = False
		self.alt_clicked = e.alt
		tool = self.tool
		if tool == SELECT_TOOL:
			self.select_tool_click(e)
		elif tool == ADD_TOOL:
			self.add_tool_click(e)
		elif tool == WIRE_TOOL:
			self.wire_tool_click(e)
		elif tool == CUT_TOOL:
			self.cut_tool_click(e)
	
	def select_tool_click(self, e):
		self.click_point = Point2(e.local)
		p = self.mouse_to_global(e)
		level = self.level
		if e.cmd:
			hit = level.find_part(p)
		else:
			hit = level.find_item(p)
		if not e.shift and hit not in level.selection:
			level.clear_selection()
		if hit:
			if e.shift:
				level.toggle_selection(hit)
			else:
				level.add_to_selection(hit)

	def add_tool_click(self, e):
		part = self.instantiate_part()
		if part:
			p = self.mouse_to_global_snapped(e)
			part.position = p
			if part.edit_params_on_create:
				part.edit_params()
			self.level.add_part(part)
	
	def wire_tool_click(self, e):
		p = self.mouse_to_global_snapped(e)
		level = self.level
		last_wp = self.editing_wire_point
		clicked_wp = not level.find_terminal(p) and level.find_wire_point(p)
		if clicked_wp and not clicked_wp.is_endpoint():
			clicked_wp = None
		if not last_wp:
			if clicked_wp and not clicked_wp.connected_part \
					and clicked_wp.wire.color_no == self.wire_color_no:
				# start extending an existing wire
				self.editing_wire_point = clicked_wp
			else:
				# start a new wire
				if clicked_wp and not clicked_wp.connected_part:
					# from the end of an existing wire of a different colour
					self.insert_junction(clicked_wp)
					p = clicked_wp.point
				wire = Wire(self.wire_color_no, p)
				self.editing_wire_point = wire.corners[0]
			self.rubber_point = None
		else:
			if last_wp is clicked_wp:
				return
			wire = last_wp.wire
			if clicked_wp and not clicked_wp.connected_part \
					and last_wp.wire is not clicked_wp.wire \
					and wire.color_no == clicked_wp.wire.color_no:
				# merge current and existing wires
				level.merge_wires(last_wp, clicked_wp)
				self.finish_wire()
			else:
				# extend current wire
				new_wp = level.extend_wire_to_point(last_wp, p)
				assert new_wp is not None
				self.editing_wire_point = new_wp
				if clicked_wp and not clicked_wp.connected_part:
					# up to the end of an existing wire of a different colour
					self.insert_junction(clicked_wp)
				if new_wp.connected_part:
					self.finish_wire()
	
	def cut_tool_click(self, e):
		p = self.mouse_to_global_snapped(e)
		action, wp = self.cut_tool_action(p)
		if action == 'cut':
			self.level.cut_wire(wp)
			if e.alt:
				self.insert_junction(wp)
		elif action == 'delete':
			self.level.delete_wire_segment(wp)
	
	def insert_junction(self, wp):
		junction = parts.Junction()
		junction.position = wp.point
		self.level.add_part(junction)
	
	def mouse_drag(self, e):
		if self.tool == SELECT_TOOL:
			p = Point2(e.local)
			d = (p - self.click_point).snap(grid_size)
			if d <> Zero2:
				self.level.offset_selection(d,
					fork_points = not self.dragged and self.alt_clicked)
				self.click_point += d
				self.dragged = True
				self.invalidate()
	
	def mouse_move(self, e):
		if self.tool == WIRE_TOOL:
			if self.editing_wire_point:
				self.rubber_point = self.mouse_to_global_snapped(e)
				self.invalidate()
	
	def key_down(self, e):
		tool = self.tool
		key = e.key
		if e.cmd:
			self.cmd_key(e)
		elif key in (K_BACKSPACE, K_DELETE):
			self.level.delete_selection()
		elif key == K_ESCAPE:
			if tool == WIRE_TOOL and self.editing_wire_point:
				self.finish_wire()
			else:
				self.shell.show_editor_menu()
		elif key == K_SPACE:
			self.set_tool(SELECT_TOOL)
		elif key == K_a:
			self.set_tool(ADD_TOOL)
		elif key == K_w:
			self.set_tool(WIRE_TOOL)
		elif key == K_d:
			self.set_tool(CUT_TOOL)
		else:
			self.pass_key(e)
	
	def cmd_key(self, e):
		key = e.key
		if key == K_h:
			self.flip_h()
		elif key == K_v:
			self.flip_v()
		elif key == K_r:
			self.rotate()
		elif key == K_e:
			self.edit_params()
		elif key == K_t:
			self.test_level()
		else:
			self.pass_key(e)
	
	def pass_key(self, e):
		self.shell.editor_menu_screen.key_down(e)
	
	def test_level(self):
		self.shell.start_level(testing = True)
	
	def finish_wire(self):
		self.editing_wire_point = None
	
	def instantiate_part(self):
		i = self.part_no
		cat = parts.catalog
		#print "Editor: instantiate_part:", i ###
		if 0 <= i < len(cat):
			return cat[i].instantiate()
	
	def mouse_to_global_snapped(self, e):
		p = self.mouse_to_global(e)
		return p.snap(grid_size)

	def get_start_time(self):
		return self.level.start_time_secs
	
	def set_start_time(self, x):
		self.level.set_start_time_secs(x)

	start_time = property(get_start_time, set_start_time)

	def get_deadline(self):
		return self.level.deadline_secs
	
	def set_deadline(self, x):
		self.level.set_deadline_secs(x)

	deadline = property(get_deadline, set_deadline)

	def get_script(self):
		return self.level.get_script()
	
	def set_script(self, x):
		self.level.set_script(x)

	script = property(get_script, set_script)

	def edit_params(self):
		self.level.edit_selection_params()

	def flip_h(self):
		self.rot_flip(2)
	
	def flip_v(self):
		self.rot_flip(4)
	
	def rotate(self):
		self.rot_flip(1)
	
	def rot_flip(self, bits):
		self.level.rotflip_selection(bits)

	def edit_time(self, caption, attr):
		form = Form()
		form.text_field('time', caption, "12:59 ")
		form['time'] = time_to_text(getattr(self.level, attr))
		form.end()
		if form.present():
			setattr(self.level, attr, text_to_time(form['time']))
			self.level.changed = True

	def edit_start_time(self):
		self.edit_time("Start Time", 'start_time_secs')
	
	def edit_deadline(self):
		self.edit_time("Deadline", 'deadline_secs')

	def edit_script(self):
		form = Form()
		form.text_field('script', "Tyger Script", 500, upper = True)
		form['script'] = self.script
		form.end()
		if form.present():
			self.script = form['script']

	def get_cursor(self, e):
		tool = self.tool
		level = self.level
		if tool == ADD_TOOL:
			return self.add_cursor
		elif tool == CUT_TOOL:
			p = self.mouse_to_global_snapped(e)
			action, wp = self.cut_tool_action(p)
			if action == 'cut':
				return self.cut_wire_cursor
			elif action == 'delete':
				return self.delete_wire_cursor
			else:
				return self.cut_cursor
		elif tool == WIRE_TOOL:
			p = self.mouse_to_global_snapped(e)
			if level.find_terminal(p):
				return self.solder_cursor
			wp = level.find_wire_point(p)
			if wp and not wp.is_connected():
				wire = self.editing_wire()
				if wire is not wp.wire:
					return self.solder_cursor
			return self.wire_cursor
	
	def cut_tool_action(self, p):
		level = self.level
		wp = level.find_wire_point(p)
		if wp:
			if wp.wire.find_end(wp) is None:
				return 'cut', wp
			else:
				return None, None
		wp = level.find_wire_segment(p)
		if wp:
			return 'delete', wp
		else:
			return None, None

	def editing_wire(self):
		wp = self.editing_wire_point
		if wp:
			return wp.wire
