#
#   5B - Distribution frame
#

import random
from pygame import Rect, draw
from controls import Form
from part import Part
from wire import get_wire_color, wire_width, num_wire_colors
from image import RotatableImage
from vector_algebra import Point2, Vector2

class DistributionFrame(Part):
	#  permutation   [int]

	color_no = 8
	num_midsections = 1
	is_rotatable = True

	block_image = RotatableImage("parts/df_terminal.png")
	corner_image = RotatableImage("parts/df_corner.png")
	middle_image = RotatableImage("parts/df_middle.png")

	block_width, block_height = block_image[0].get_size()
	block_height -= 2 # The blocks overlap a little
	middle_width, end_height = middle_image[0].get_size()

	def __init__(self):
		self.terminal_points = []
		self.permutation = []
		Part.__init__(self)
	
	def init_from_catalog(self, num_ways):
		self.set_num_ways(num_ways)
	
	def num_ways(self):
		return len(self.terminal_points) // 2
	
	def get_oriented_rect(self):
		n = self.num_ways()
		w = 2 * self.block_width + self.num_midsections * self.middle_width
		h = n * self.block_height + 2 * self.end_height
		if self.orientation & 1:
			w, h = h, w
		r = Rect((0, 0, w, h))
		r.center = (0, 0)
		return r

	def draw(self, surf, t):
		tt = t * self.local_to_global_transformation()
		n = self.num_ways()
		bw = self.block_width
		bh = self.block_height
		mw = self.middle_width
		eh = self.end_height
		o1 = self.orientation
		nm = self.num_midsections
		if o1 & 1:
			fh = 4    # flip horizontal
			fv = 2    # flip vertical
		else:
			fh = 2
			fv = 4
		x = (bw + nm * mw) // 2       # vertical midline of block column
		y = (n * bh + eh) // 2   # horizontal midline of end section
		self.draw_image(surf, tt, self.corner_image, o1,     (-x, -y))
		self.draw_image(surf, tt, self.corner_image, o1 ^ fh, (x, -y))
		self.draw_image(surf, tt, self.corner_image, o1 ^ fv, (-x, y))
		self.draw_image(surf, tt, self.corner_image, o1 ^ 6, (x, y))
		for i in xrange(1, nm + 1):
			ex = x - i * mw
			self.draw_image(surf, tt, self.middle_image, o1,     (ex, -y))
			self.draw_image(surf, tt, self.middle_image, o1 ^ fv, (ex, y))
		bi = self.block_image
		y0 = - (((n - 1) * bh) // 2)
		for i in xrange(n):
			y = y0 + i * bh
			self.draw_image(surf, tt, bi, o1, (-x, y))
			self.draw_image(surf, tt, bi, o1 ^ 6, (x, y))
		sx = x - 20
		perm = self.permutation
		color = get_wire_color(self.color_no)
		width = wire_width
		for i in xrange(n):
			y1 = y0 + i * bh
			y2 = y0 + perm[i] * bh
			p1 = tt * Point2(-sx, y1)
			p2 = tt * Point2(sx, y2)
			draw.line(surf, color, p1.coords, p2.coords, width)
	
	def draw_image(self, surf, t, image, orientation, center):
		p = t * Point2(center)
		image.draw(surf, p, orientation)
	
	def set_num_ways(self, num_ways):
		tp = []
		bw = self.block_width
		bh = self.block_height
		mw = self.middle_width
		nm = self.num_midsections
		x = (nm * mw) // 2 + bw - 10
		y0 = - ((bh * (num_ways - 1)) // 2)
		for i in xrange(num_ways):
			y = y0 + bh * i
			tp.append((-x, y))
			tp.append((x, y))
		self.set_terminal_points(tp)
		if len(self.permutation) <> num_ways:
			self.permutation = range(num_ways)
	
	def update_terminal_points(self):
		self.set_num_ways(self.num_ways())

	def reset(self):
		#print "DF: reset" ###
		random.shuffle(self.permutation)

	def get_params(self, params):
		params.append(self.num_ways())
		params.append(self.color_no)
		params.append(self.num_midsections)
	
	def read_params(self, r):
		num_ways = r.get_arg(int)
		self.color_no = r.get_arg(int, DistributionFrame.color_no)
		self.num_midsections = r.get_arg(int, 1)
		self.set_num_ways(num_ways) # Do last to update terminal points

	def node_changed(self, i):
		perm = self.permutation
		way1 = i >> 1
		side = i & 1
		if side:
			way2 = perm.index(way1)
		else:
			way2 = perm[way1]
		j = way2 << 1 | (side ^ 1)
		self.propagate(i, j)

	def edit_params(self):
		form = Form()
		form.int_field('ways', "No. of ways")
		form.int_field('color', "Colour no.")
		form.int_field('width', "Width")
		form.end()
		form['ways'] = self.num_ways()
		form['color'] = self.color_no
		form['width'] = self.num_midsections
		if form.present():
			self.color_no = max(0, min(form['color'], num_wire_colors() - 1))
			self.num_midsections = max(1, form['width'])
			self.set_num_ways(form['ways']) # Do last, to update terminal points
			return True
		else:
			return False
